#!/usr/bin/perl

#
# decode.pl
#
# (C) Justin Schoeman 2013
#
# Very crude attempt to read MGL airtalk messages from a digital audio recording of the message bus.
# Should work on any recording 44100 samples per second, or higher (only tested at 48000 samples per 
# second so far.
#
# Currently the only message frames supported are Ultra Horizon type 1 subtypes 1-3. Others should
# be easy to add.
#
# The one MAJOR todo is autoscaling the edge threshold.  The current code will only work with
# saturated recordings.
#
# Connect to Airtlak bus through the microphone cable:
#
#   MIC o------------------------------WWWWW---------------o Airtalk
#              |    |    |              100k
#              |    |    |              R1
#              Z    _    _
#       R2 10k Z    ^    V
#              Z    _    _
#              |    | D1 | D2
#              |    |    |
#  GND o----------------------------------------------------o GND
#
# R1 and R2 form a voltage divider and current limiter
# D1 and D2 form a clamp and protection circuit.
#



use strict;
use Audio::Wav;
use Data::Dumper;
use IO::File;
use Audio::OSS qw(:funcs :formats);

#print STDERR "ver: $]\n";
#exit -1;

sub usage()
{
  print STDERR "Usage: $0 [-render] [-oss] <audio filename|oss device>\n";
  exit -1;
}

my $fname;
my $swrender = 0;
my $ossmode = 0;
foreach my $i ( @ARGV ) {
  if($i =~ /^-(.*)$/) {
    # switch
    if($1 eq "render") {
      print STDERR "ENABLE RENDERING\n";
      $swrender = 1;
    } elsif($1 eq "oss") {
      print STDERR "ENABLE OSS MODE\n";
      $ossmode = 1;
    } else {
      print STDERR "UNHANDLED SWITCH '$1'\n";
      usage();
    }
  } else {
    # filename
    if($fname) {
      print STDERR "MULTIPLE FILE NAMES SPECIFIED!\n";
      usage();
    }
    $fname = $i;
  }
}

#
# open device
#
my $oss_dsp;
my $wavrd;
my $sr;
my $smax;
if($ossmode) {
  #
  # oss device
  #
  
  if(!$fname) {
    $fname = "/dev/dsp";
  }
  print STDERR "Use OSS device '$fname'.\n";
  $oss_dsp = new IO::File;
  $oss_dsp->open("<$fname") or die "OSS open failed: $!";
  dsp_reset($oss_dsp) or die "OSS reset failed: $!";
  my $mask = get_supported_fmts($oss_dsp);
  print STDERR "OSS suppoorted formats " . sprintf "%08X\n", $mask;
  if($mask & AFMT_S8) {
    set_fmt($oss_dsp, AFMT_S8) or die "OSS set format failed: $!";
    $smax = 256;
  } else {
    print STDERR "OSS AFMT_S8 NOT SUPPORTED!\n";
    exit -1;
  }
  $sr = set_sps($oss_dsp, 48000);
  print STDERR "OSS rate: $sr\n";
  set_fragment($oss_dsp, 12, 32767) || die "OSS set fragment failed: $!";
} else {
  #
  # wav file
  #
  
  if(!$fname) {
    print STDERR "No filename specified!\n";
    usage();
  }
  if(!-r $fname) {
    print STDERR "File '$fname' does not exist, or is not readable.\n";
    exit -1;
  }
  my $wav = new Audio::Wav;
  if(!$wav) {
    print STDERR "Error instantiating new WAV reader...\n";
    exit -1;
  }
  $wavrd = $wav->read($fname);
  if(!$wavrd) {
    print STDERR "Error reading WAV file '$fname': $!\n";
    exit -1;
  }
  my $wavdt = $wavrd->details();
  print STDERR "Details: " . Dumper($wavdt);
  
  if($wavdt->{'channels'} != 1) {
    print STDERR "only support 1 channel for now...\n";
    exit 0;
  }
  $sr = $wavdt->{'sample_rate'};
  print STDERR "SR: $sr\n";
  # data range
  $smax = 2 ** $wavdt->{'bits_sample'};
  print STDERR "SMAX: $smax\n";

  #my $startpos = $wavrd->position_samples();
  #print STDERR "STARTPOS: $startpos\n";
}
#exit 0;

# stream bit rate
my $br = 19200.0;
print STDERR "BR: $br\n";
# bit period
my $bp = 1.0 / $br;
print STDERR "BP: $bp\n";
# sample period
my $sp = 1.0 / $sr;
print STDERR "SP: $sp\n";


#exit 0;

# gack - manually pre-process amplitude
# FIXME FIXME FIXME - only works for saturated input - need to pre-process stream to get a good scale factor
my $amp = $smax/8;

# output data
my %data;

# absolute clock
my $aclock = 0;

#
# frame type definitions
#
# descr = human readable frame type description
# seq = sequence ordering for output data
# fields = sequential list of fields in frame
#  field - key to use to record in 'data' (undef to not store data)
#  type - perl 'pack' type to use to unpack field
#  pproc - post processing command - opearte on variable '$j'
#

my %ftypes = (
  '1-0' => {
    'descr' => 'flight data',
    'seq' => 1,
    'fields' => [
      # lsssCsCCsC
      {'field' => 'altitude', 'type' => 'l'},
      {'field' => 'qnh', 'type' => 'S'},
      {'field' => 'vsi', 'type' => 's'},
      {'field' => 'asi', 'type' => 'S'},
      {'field' => 'ambientT', 'type' => 'C', 'pproc' => '$j = ($j*10/82)'},
      {'field' => 'flighttime_minutes', 'type' => 'C'},
      {'field' => 'flighttime_seconds', 'type' => 'C'},
      {'field' => 'in_flight', 'type' => 'C'},
      {'field' => 'rdac_fail', 'type' => 'C'},
      {'field' => 'baro', 'type' => 'S'},
      {'field' => undef, 'type' => 'C'},
    ]
  },
  '1-1' => {
    'descr' => 'temperature data',
    'seq' => 2,
    'fields' => [
      # ssssssssssssC
      {'field' => 'tc1', 'type' => 'S'},
      {'field' => 'tc2', 'type' => 'S'},
      {'field' => 'tc3', 'type' => 'S'},
      {'field' => 'tc4', 'type' => 'S'},
      {'field' => 'tc5', 'type' => 'S'},
      {'field' => 'tc6', 'type' => 'S'},
      {'field' => 'tc7', 'type' => 'S'},
      {'field' => 'tc8', 'type' => 'S'},
      {'field' => 'tc9', 'type' => 'S'},
      {'field' => 'tc10', 'type' => 'S'},
      {'field' => 'tc11', 'type' => 'S'},
      {'field' => 'tc12', 'type' => 'S'},
      {'field' => undef, 'type' => 'C'},
    ]
  },
  '1-2' => {
    'descr' => 'engine data',
    'seq' => 3,
    'fields' => [
      # ssss sssC sCss sssC
      {'field' => 'oilT', 'type' => 'S'},
      {'field' => 'oilP', 'type' => 'S', 'pproc' => '$j = $j * 0.1'},
      {'field' => 'waterT', 'type' => 'S'},
      {'field' => 'cht1', 'type' => 'S'},
      {'field' => 'cht2', 'type' => 'S'},
      {'field' => 'rpm', 'type' => 'S'},
      {'field' => 'man', 'type' => 'S', 'pproc' => '$j = $j / 1000.0 * 29.53'},
      {'field' => 'fuelP', 'type' => 'C', 'pproc' => '$j = $j * 0.1'},
      {'field' => 'rotor', 'type' => 'S'},
      {'field' => 'voltage', 'type' => 'C', 'pproc' => '$j = $j * 0.1'},
      {'field' => 'amps', 'type' => 's'},
      {'field' => 'fuelflow', 'type' => 'S'},
      {'field' => 'fuellevel1', 'type' => 'S'},
      {'field' => 'fuellevel2', 'type' => 'S'},
      {'field' => 'fuellevelcalc', 'type' => 'S'},
      {'field' => undef, 'type' => 'C'},
    ]
  },
);
#print STDERR "ft: " . Dumper(\%ftypes);

# Preprocess frames to determine
#  template - perl 'unpack' template to decode entire frame
#  len - number of bytes that should be in the frame
#  cnt - number of data fields that should be in the frame
#
# UGLY - mangles the 'normal' unpack template into force intel endianness...
#
foreach my $i (keys %ftypes) {
  print STDERR "preproc frame '$i'...\n";
  $ftypes{$i}->{'template'} = "";
  $ftypes{$i}->{'len'} = 0;
  $ftypes{$i}->{'cnt'} = 0;
  foreach my $j (@{$ftypes{$i}->{'fields'}}) {
    print STDERR "field " . Dumper($j);
    if(!defined $j->{'type'}) {
      print STDERR "Frame '$i' has no type for field " . Dumper($j);
      exit 0;
    }
    if($j->{'type'} eq 'l') {
      # longint = signed 32 bit
      $ftypes{$i}->{'len'} += 4;
      if($] >= 5.014) {
        $ftypes{$i}->{'template'} .= "l<";
      } else {
        print STDERR "WARNING - PERL TOO OLD - WILL ONLY WORK ON BE MACHINES\n";
        $ftypes{$i}->{'template'} .= "l";
      }
    } elsif($j->{'type'} eq 's') {
      # integer/word = signed 16 bit
      $ftypes{$i}->{'len'} += 2;
      if($] >= 5.014) {
        $ftypes{$i}->{'template'} .= "s<";
      } else {
        print STDERR "WARNING - PERL TOO OLD - WILL ONLY WORK ON BE MACHINES\n";
        $ftypes{$i}->{'template'} .= "s";
      }
    } elsif($j->{'type'} eq 'S') {
      # word = unsigned 16 bit
      $ftypes{$i}->{'len'} += 2;
      if($] >= 5.014) {
        $ftypes{$i}->{'template'} .= "S<";
      } else {
        print STDERR "WARNING - PERL TOO OLD - WILL ONLY WORK ON BE MACHINES\n";
        $ftypes{$i}->{'template'} .= "S";
      }
    } elsif($j->{'type'} eq 'C') {
      # byte = unsigned 8 bit
      $ftypes{$i}->{'len'} += 1;
      $ftypes{$i}->{'template'} .= "C";
    } else {
      print STDERR "Frame '$i' has no unsupported type '$j->{'type'}' for field " . Dumper($j);
      exit 0;
    }
    $ftypes{$i}->{'cnt'} += 1;
  }
  print STDERR "template: '$ftypes{$i}->{'template'}'\n";
  print STDERR "length: '$ftypes{$i}->{'len'}'\n";
  print STDERR "count: '$ftypes{$i}->{'cnt'}'\n";
}
#exit 0;

#
# Process received bytes into frames, then process complete frames
#

# frame state
my @frame;

# pre-print column headers...
print "seconds,";
foreach my $i ( sort {$ftypes{$a}->{'seq'} <=> $ftypes{$b}->{'seq'}} keys %ftypes ) {
  print STDERR "ftype '$i'\n";
  foreach my $j ( @{$ftypes{$i}->{'fields'}} ) {
    print STDERR "field '$j->{'field'}'\n";
    if(defined $j->{'field'}) {
      print "$j->{'field'},";
    }
  }
}
print "\n";
#exit 0;


# add a byte to a frame
# if 'undef' is received, then the serial data is interrupted - assume it is a frame break
sub do_byte($) 
{
  my ($byte) = @_;
  if(defined $byte) {
    if($swrender) { print STDERR "BYTE: $byte\n"; }
    # not a frame break - just add the byte to the frame buffer...
    push @frame, $byte;
    return;
  }
  # break - process frame
  # possibly multiple frames and/or loss of sync - do inside a while loop, until the entire frame buffer is consumed
  while(@frame) {
    if($swrender) { print STDERR "BREAK\n"; }
    #print STDERR "got frame: " . Dumper(\@frame);
    # calculate checksum on the fly...
    my $cks = 0xa5;
    # STX
    my $stx = shift @frame;
    print STDERR "stx: $stx\n";
    if($stx != 130) {
      print STDERR "STX $stx != 130...\n";
      next;
    }
    # target address
    my $addr = shift @frame;
    $cks ^= $addr;
    print STDERR "addr: $addr\n";
    if($addr != 255) {
      print STDERR "Target address $addr != 255...\n";
      next;
    }
    # data length
    my $len = shift @frame;
    $cks ^= $len;
    print STDERR "len: $len\n";
    # type
    my $type = shift @frame;
    $cks ^= $type;
    print STDERR "type: $type\n";
    # subtype
    my $stype = shift @frame;
    $cks ^= $stype;
    print STDERR "stype: $stype\n";
    # spool out sub frame data
    my $sframe;
    while($len > 0) {
      my $sb = shift @frame;
      if(!defined $sb) {
        last;
      }
      if($len == 2) {
        # 2nd last byte is checksum
        print STDERR "cks: $cks $sb\n";
        if($cks != $sb) {
          print STDERR "CHECKSUM FAIL ($sb != $cks)!\n";
          #exit 0;
          last;
        }
      } elsif($len == 1) {
        # last byte is ETX
        print STDERR "etx: $sb\n";
        if($sb != 131) {
          print STDERR "ETX $sb != 131...\n";
          #exit 0;
          last;
        }
      } else {
        # all others are normal data bytes
        $cks ^= $sb;
        $sframe .= chr($sb);
      }
      $len--;
    }
    if($len) {
      # 'last' before munching all data = some sort of error, or just plain ran out of data....
      print STDERR "FRAME ERROR, OR RAN OUT OF DATA FOR LEN '$len'...\n";
      #exit 0;
      next;
    }
    #print STDERR "sframe: " . Dumper(\$sframe);
    # look up frame type definition
    my $f = $ftypes{"$type-$stype"};
    if(!defined $f) {
      print STDERR "No frame handler for frame type '$type' subtype '$stype'.\n";
      #exit 0;
      next;
    }
    print STDERR "Got frame '$f->{'descr'}'...\n";
    # verify we have ther correct data for the template
    if(length($sframe) != $f->{'len'}) {
      print STDERR "$f->{'descr'} message ($type-$stype) length is " . length($sframe) . " not $f->{'len'}!\n";
      #exit 0;
      next;
    }
    # unpack the frame
    my @p = unpack($f->{'template'}, $sframe);
    #push @p, "foo";
    # check that we got all the elements
    if(@p != $f->{'cnt'}) {
      print STDERR "$f->{'descr'} message ($type-$stype) decoded to " . (0 + @p) . " parts not $f->{'cnt'}!\n";
      exit 0;
    }
    #print STDERR "P: " . Dumper(\@p);
    # process each element
    foreach my $i ( @{$f->{'fields'}}) {
      #print STDERR "F: " . Dumper($i);
      my $j = shift @p;
      #print STDERR "Value: '$j'\n";
      # post-process, if required
      if(defined $i->{'pproc'}) {
        eval "$i->{'pproc'}";
        print STDERR "Postproc value: '$j'\n";
      }
      # record into data structure
      if(defined $i->{'field'}) {
        $data{$i->{'field'}} = sprintf "%.1f", $j;
      }
    }
    # output data
    # use etype 1-0 to trigger this, as it is the most frequent frame...
    if($type eq "1" && $stype eq "0") {
      #print STDERR "DATA: " . Dumper(\%data);
      my $o = sprintf "%.1f,", $aclock;
      foreach my $i ( sort {$ftypes{$a}->{'seq'} <=> $ftypes{$b}->{'seq'}} keys %ftypes ) {
        #print STDERR "ftype '$i'\n";
        foreach my $j ( @{$ftypes{$i}->{'fields'}} ) {
          #print STDERR "field '$j->{'field'}'\n";
          if(defined $j->{'field'}) {
            $o .= "$data{$j->{'field'}},";
          }
        }
      }
      print "$o\n";
    }
    #exit 0;
  }
}

#
# state engine to handle received bits
#

# bit state
# rs - reset sync
# as - waiting for sync
# br - break received
# sb - received start bit - run out bits
# sb - receiped stop bit
my $bs = "rs";
my $bc = 0;
my $byte;
sub do_bit($)
{
  my ($bit) = @_;
  #print STDERR "$bit\n";
  if($bs eq "rs") {
    # reset to sync wait
    print STDERR "RESET!\n";
    # go to async state, and reset break count
    $bs = "as";
    $bc = 0;
    # send a 'break' to flush any partial frame data...
    do_byte(undef);
    # fall through, and process the received bit
  }
  if($bs eq "as") {
    # wait for break (15 sequential zeors)
    if($bit) {
      # got '1' while waiting for break
      if($bc > 0) {
        # abuse bc, so we don't repeat this error for a stream of ones...
        print STDERR "Received '1' while waiting for break - reset break count...\n";
        $bc = 0;
      }
      return;
    }
    # got a zero? count to 15
    $bc++;
    if($bc > 15) {
      print STDERR "Break received...\n";
      do_byte(undef);
      $bs = "br";
    }
    return;
  }
  if($bs eq "br" || $bs eq "st") {
    # received break/byte - wait for start bit
    if(!$bit) { 
      # zero...
      if($bs eq "st") {
        # waiting for start nit - check for break...
        # abuse bc to prevent repeating 'break' transmissions...
        if($bc < 100) {
          $bc++;
          if($bc > 15) {
            # 15 sequential zeros...
            print STDERR "BREAK!\n\n";
            do_byte(undef);
            # no more break checking...
            $bc = 100;
          }
        }
      }
      return;
    }
    # '1' - got start bit
    #print STDERR "Got start bit...\n";
    $bs = "sb";
    # bc will be used as the bit value now...
    $bc = 1;
    $byte = 0;
    return;
  }
  if($bs eq "sb") {
    # receiving bits
    if($bc < 256) {
      # received bit
      # sense seems to be inverted???
      if(!$bit) { $byte += $bc; }
      # next bit value
      $bc *= 2;;
      return;
    }
    # stop bit - should be '0'...
    #print STDERR "byte '$byte' stop '$b'\n";
    if($b) {
      print STDERR "got stop bit '1' - something is wrong...\n";
      #exit 0;
      my $bs = "rs";
      return;
    }
    # handle byte, then wait for next start bit
    #print STDERR "GOT BYTE '$byte'\n";
    do_byte($byte);
    $bs = "st";
    $bc = 0;
    return;
  }
  print STDERR "Unhandled state '$bs'.\n";
  exit -1;
}

#
# Sound file parsing - detect edges, and clock events
#

# current bit state
my $state = 0;
# last analog value
my $lastd = 0;
# bit clock
my $bc = 0;
# bit value (display/debug purposes
my $bit = " ";
# test/debug exit counter
my $exitcnt = 0;
my @oss_samples;
my $oss_buffer;
my $oss_buffer_sz = 4096;
my @oss_frags;
my $d;
while() {
  # tick absolute clock
  $aclock += $sp;
  if($oss_dsp) {
    #print STDERR "OSSSAMPLES " . 0+@oss_samples . "\n";
    if(!@oss_samples) {
      @oss_frags = get_inbuf_info($oss_dsp);
      #print STDERR Dumper(\@oss_frags);
      if($oss_frags[0] < 1) {
        # at least one frag (sleep if necessary)
        $oss_frags[0] = $oss_buffer_sz;
      } else {
        # otherwise, all complete frags
        $oss_frags[0] *= $oss_buffer_sz;
      }
      $oss_dsp->read($oss_buffer, $oss_frags[0]);
      if(length($oss_buffer) != $oss_frags[0]) {
        print STDERR "Error reading from OSS device ($!)...\n";
        exit -1;
      }
      push @oss_samples, unpack("c" . $oss_frags[0], $oss_buffer);
      print STDERR "samples: " . (0+@oss_samples) . "\n";
      if(@oss_samples != $oss_frags[0]) {
        print STDERR "Error decoding OSS data ($!)...\n";
        exit -1;
      }
    }
    $d = shift @oss_samples;
  } else {
    my @c = $wavrd->read();
    if(!@c) {
      print STDERR "EOF!\n";
      last;
    }
    #print STDERR "s1: " . Dumper(\@c);
    # get next sample (channel 0, i hope...)
    # FIXME - configurable channel!
    $d = $c[0];
    #print STDERR "$d\n";
  }
  # state transition detected flag
  my $st = 0;
  if($state) {
    # was one, look for a large downward step
    if($d < $lastd - $amp) {
      # change to '0' state
      $state = 0;
      $st = 1;
    }
  } else {
    # was zero - look for large upward step
    if($d > $lastd + $amp) {
      # change to '1' state
      $state = 1;
      $st = 1;
    }
  }
  # record last value for use in next step calculation
  $lastd = $d;
  $bit = " ";
  if($st) {
    # state transition - set bit, and reset bitclock
    $bit = $state;
    do_bit($state);
    # hmmm... this doesn't work too well, no idea why LE sampling seems to work so much better?
    # FIXME - configurable fudge factor? Or is it needed? '0' seems to work very well...
    #$bc = -$bp/4; # pretend we actually sampled 1/4 into the future (well after the edge)
    #$bc = -$bp/8; # pretend we actually sampled 1/4 into the future (well after the edge)
    $bc = 0;
  } else {
    # no state transition - tick the bitclock by one sample period, and see if a whole bit period has passed
    # tick the bit clock
    $bc += $sp;
    if($bc >= $bp) {
      # bit period passed - sample now
      $bit = $state;
      do_bit($state);
      # deduct bit period to prevent decreasing accuracy over time
      $bc -= $bp;
    }
  }
  if(!$swrender) { next; }
  # test code...
  #if($d > 0) {
  #  $exitcnt++;
  #  if($exitcnt > 400) { exit 0; }
  #}
  my $val = $d;
  $val += $smax / 2;
  $val /= $smax;
  $val *= 70;
  #$d /=4;
  #$d += 32;
  my $o = sprintf "%05d", $d;
  $o .= "|$bit|$state|";
  for(my $i = 0; $i < $val; $i++) { $o .= " "; }
  #print STDERR "# ($dc $lmin $lmax)\n";
  #print STDERR "$o*\n";
  print STDERR "$o* ($d)\n";
  #exit 0;
}
